@@ -6,20 +6,20 @@ import threading
 import time
 from collections import defaultdict
 from concurrent.futures import ThreadPoolExecutor
-from dataclasses import dataclass, field
+from dataclasses import dataclass
 from datetime import datetime
 from threading import Thread
 from typing import Optional, Dict, List, Any, cast, Tuple
 from urllib.parse import urlparse, urljoin
 
-from isisdl.backend.downloads import SessionWithKey, MediaType, MediaContainer, Status, DownloadThrottler, PreStatus, PreStatusInfo
-from isisdl.backend.utils import User, path, sanitize_name, args, on_kill, database_helper, _course_downloader_transformation
-from isisdl.settings import course_dir_location, enable_multithread, checksum_algorithm, is_testing, checksum_num_bytes
+from isisdl.backend.downloads import SessionWithKey, MediaType, MediaContainer, DownloadStatus, DownloadThrottler, InfoStatus, PreStatusInfo
+from isisdl.backend.utils import User, path, sanitize_name, args, on_kill, database_helper, config, generate_error_message
+from isisdl.settings import enable_multithread, checksum_algorithm, video_size_discover_num_threads
 
 
 @dataclass
 class PreMediaContainer:
-    name: str
+    _name: str
     file_id: str
     url: str
     location: str
@@ -53,31 +53,24 @@ class PreMediaContainer:
         time = datetime.fromtimestamp(last_modified)
 
         if "webservice/pluginfile.php" not in url and "mod/videoservice/file.php" not in url:
-            # TODO: Server
+            # Later: Server
             pass
 
         return cls(name, file_id, sanitized_url, location, time, course.course_id, is_video, size)
 
+    @property
+    def path(self) -> str:
+        return os.path.join(self.location, sanitize_name(self._name))
+
     def dump(self) -> bool:
         assert self.checksum is not None
         return database_helper.add_pre_container(self)
 
-    def calculate_online_checksum(self, s: SessionWithKey) -> Tuple[str, int]:
-        while True:
-            running_download = s.get_(self.url, params={"token": s.token}, stream=True)
+    def __str__(self) -> str:
+        return sanitize_name(self._name)
 
-            if running_download is None:
-                continue
-
-            if not running_download.ok:
-                assert False
-
-            break
-
-        chunk = running_download.raw.read(checksum_num_bytes, decode_content=True)
-        size = len(chunk)
-
-        return checksum_algorithm(chunk + str(size).encode()).hexdigest(), size
+    def __repr__(self) -> str:
+        return self.__str__()
 
     def __hash__(self) -> int:
         return self.file_id.__hash__()
@@ -95,12 +88,15 @@ class Course:
 
         return cls(name, id)
 
-    def __post_init__(self) -> None:
+    def __post_init(self) -> None:
+        self.make_directories()
+
+    def make_directories(self) -> None:
         for item in MediaType.list_dirs():
             os.makedirs(self.path(item), exist_ok=True)
 
     def download_videos(self, s: SessionWithKey) -> List[PreMediaContainer]:
-        if args.disable_videos:
+        if args.disable_videos or not config.download_videos:
             return []
 
         url = "https://isis.tu-berlin.de/lib/ajax/service.php"
@@ -120,8 +116,35 @@ class Course:
         if videos_json["error"]:
             return []
 
-        return [PreMediaContainer.from_course(item["title"].strip() + item["fileext"], item["id"], item["url"], self, item["date"], size=item["duration"])
-                for item in videos_json["data"]["videos"]]
+        videos_json = videos_json["data"]["videos"]
+
+        def find_out_size(info: Dict[str, str]) -> Optional[int]:
+            req = s.head_(info["url"], allow_redirects=True)
+            if req is None:
+                return None
+
+            if "content-length" not in req.headers:
+                return None
+
+            try:
+                size = int(req.headers["content-length"])
+                req.close()
+                return size
+
+            except ValueError:
+                return None
+
+        video_cache = database_helper.get_video_cache(self.name)
+        not_found_videos = [item for item in videos_json if item["url"] not in video_cache]
+
+        with ThreadPoolExecutor(video_size_discover_num_threads) as ex:
+            sizes = list(ex.map(find_out_size, not_found_videos))
+
+        video_cache.update({video["url"]: size for video, size in zip(videos_json, sizes)})
+        database_helper.set_video_cache(video_cache, self.name)
+
+        return [PreMediaContainer.from_course(item["title"].strip() + item["fileext"], item["id"], item["url"], self, item["date"], size=video_cache.get(item["url"], -1))
+                for item in videos_json]
 
     def download_documents(self, helper: RequestHelper) -> List[PreMediaContainer]:
         if args.disable_documents:
@@ -154,11 +177,11 @@ class Course:
                     continue
 
                 if re.match(".*mod/(?:folder|resource)/.*", url) is None:
-                    # TODO: server
+                    # Later: Server
                     pass
 
                 if "contents" not in file:
-                    # TODO: Server
+                    # Later: Server
                     continue
 
                 prev_len = len(all_content)
@@ -172,23 +195,30 @@ class Course:
                     }
 
                     if url not in known_bad_urls:
-                        # TODO: Server
+                        # Later: Server
                         pass
 
         return all_content
 
     def path(self, *args: str) -> str:
-        """
-        Custom path function that prepends the args with the `download_dir` and course name.
-        """
-        return path(course_dir_location, sanitize_name(self.name), *args)
+        # Custom path function that prepends the args with the course name.
+        return str(path(sanitize_name(self.name), *args))
 
     @property
     def ok(self) -> bool:
-        if args.whitelist != [True] and self in args.whitelist:
+        if config.whitelist is None and config.blacklist is None:
             return True
 
-        return self in args.whitelist and self not in args.blacklist
+        if config.whitelist is None and config.blacklist is not None:
+            return self.course_id not in config.blacklist
+
+        if config.whitelist is not None and config.blacklist is None:
+            return self.course_id in config.whitelist
+
+        if config.whitelist is not None and config.blacklist is not None:
+            return self.course_id in config.whitelist and self.course_id not in config.blacklist
+
+        assert False
 
     def __str__(self) -> str:
         return self.name
@@ -211,45 +241,63 @@ class Course:
     def __hash__(self) -> int:
         return self.course_id
 
+    def __lt__(self, other: Course) -> bool:
+        return self.course_id < other.course_id
+
 
-@dataclass
 class RequestHelper:
     user: User
-    session: SessionWithKey = field(init=False)
-    courses: List[Course] = field(default_factory=lambda: [])
-    course_id_mapping: Dict[int, Course] = field(default_factory=lambda: {})
-    _meta_info: Dict[str, str] = field(default_factory=lambda: {})
+    session: SessionWithKey
+    courses: List[Course]
+    course_id_mapping: Dict[int, Course]
+    _meta_info: Dict[str, str]
+    _instance: Optional[RequestHelper] = None
+    _instance_init: bool = False
 
-    def __post_init__(self) -> None:
+    def __init__(self, user: User):
+        if self._instance_init:
+            return
 
-        temp = SessionWithKey.from_scratch(self.user, pre_status)
-        if temp is None:
-            print(f"I had a problem getting the {self.user = !s}. You have probably entered the wrong credentials.\nBailing out…")
-            exit(42)
+        self.user = user
+        session = SessionWithKey.from_scratch(self.user, pre_status)
+        if session is None:
+            print(f"I had a problem getting the user {self.user}. You have probably entered the wrong credentials.\nBailing out…")
+            exit(1)
 
-        self.session = temp
+        self.session = session
+        self.courses = []
+        self.course_id_mapping = {}
 
-        pre_status.set_status(PreStatusInfo.get_info)
         self._meta_info = cast(Dict[str, str], self.post_REST("core_webservice_get_site_info"))
 
-        pre_status.set_status(PreStatusInfo.get_courses)
-        self._get_courses()
+        self.get_courses()
 
-        if args.verbose:
-            print("I am downloading the following courses:\n" + "\n".join(item.name for item in self.courses))
+        RequestHelper._instance_init = True
 
-    def _get_courses(self) -> None:
+    def __new__(cls, user: User) -> RequestHelper:
+        if RequestHelper._instance is None:
+            RequestHelper._instance = super().__new__(cls)
+
+        return RequestHelper._instance
+
+    def make_course_paths(self) -> None:
+        for course in self.courses:
+            if not os.path.exists(course.path()):
+                os.makedirs(course.path(), exist_ok=True)
+            course.make_directories()
+
+    def get_courses(self) -> None:
         res = cast(List[Dict[str, str]], self.post_REST("core_enrol_get_users_courses", {"userid": self.userid}))
+        courses = []
         for item in res:
             course = Course.from_dict(item)
-
             self.course_id_mapping.update({course.course_id: course})
             database_helper.add_course(course)
 
             if course.ok:
-                self.courses.append(course)
-                if not os.path.exists(course.path()):
-                    os.makedirs(course.path(), exist_ok=True)
+                courses.append(course)
+
+        self.courses = sorted(courses, reverse=True)
 
     def post_REST(self, function: str, data: Optional[Dict[str, Any]] = None, params: Optional[Dict[str, Any]] = None) -> Optional[Any]:
         data = data or {}
@@ -259,7 +307,7 @@ class RequestHelper:
             "moodlewssettingfileurl": "true",
             "moodlewsrestformat": "json",
             "wsfunction": function,
-            "wstoken": self.session.token
+            "wstoken": self.session.token,
         })
 
         url = "https://isis.tu-berlin.de/webservice/rest/server.php"
@@ -272,28 +320,43 @@ class RequestHelper:
         return response.json()
 
     def download_content(self) -> List[PreMediaContainer]:
+        exception_lock = threading.Lock()
+
         pre_status.set_status(PreStatusInfo.content)
 
         # The number of threads that are going to spawn
         pre_status.set_max_content(len(self.courses) * 2 + 1)
 
         def download_videos(ret: List[PreMediaContainer], course: Course, session: SessionWithKey) -> None:
-            ret.extend(course.download_videos(session))
+            try:
+                ret.extend(course.download_videos(session))
 
-            assert pre_status.status == PreStatusInfo.content
-            pre_status.status.value.append(None)
+                assert pre_status.status == PreStatusInfo.content
+                pre_status.status.value.append(None)
+            except Exception:
+                with exception_lock:
+                    generate_error_message()
 
         def download_documents(ret: List[PreMediaContainer], course: Course) -> None:
-            ret.extend(course.download_documents(self))
+            try:
+                ret.extend(course.download_documents(self))
 
-            assert pre_status.status == PreStatusInfo.content
-            pre_status.status.value.append(None)
+                assert pre_status.status == PreStatusInfo.content
+                pre_status.status.value.append(None)
+            except Exception:
+                with exception_lock:
+                    generate_error_message()
 
         def download_mod_assign(ret: List[PreMediaContainer]) -> None:
-            ret.extend(self.download_mod_assign())
+            try:
+                ret.extend(self.download_mod_assign())
 
-            assert pre_status.status == PreStatusInfo.content
-            pre_status.status.value.append(None)
+                assert pre_status.status == PreStatusInfo.content
+                pre_status.status.value.append(None)
+
+            except Exception:
+                with exception_lock:
+                    generate_error_message()
 
         if enable_multithread:
             collect: List[Tuple[List[PreMediaContainer], List[PreMediaContainer]]] = [([], []) for _ in range(len(self.courses))]
@@ -333,7 +396,10 @@ class RequestHelper:
             return sorted(lst, key=lambda x: x.time, reverse=True)
 
         # Download the newest files first
-        return sort(documents) + sort(mod_assign) + sort(videos)
+        all_files = sort(documents + mod_assign) + sort(videos)
+        check_for_conflicts_in_files(all_files)
+
+        return all_files
 
     def download_mod_assign(self) -> List[PreMediaContainer]:
         if args.disable_documents:
@@ -354,7 +420,6 @@ class RequestHelper:
                         file["filepath"] = assignment["name"]
                         all_content.append(PreMediaContainer.from_api(file, assignment["id"], self.course_id_mapping[course["id"]]))
 
-        # TODO: Also get the submission contents.
         return all_content
 
     @property
@@ -365,7 +430,7 @@ class RequestHelper:
 def check_for_conflicts_in_files(files: List[PreMediaContainer]) -> None:
     conflicts = defaultdict(list)
     for item in files:
-        conflicts[item.name].append(item)
+        conflicts[item._name].append(item)
 
     items: List[List[PreMediaContainer]] = [sorted(item, key=lambda x: x.time if x.time is not None else -1) for item in conflicts.values() if len(item) != 1]
     for row in items:
@@ -375,14 +440,12 @@ def check_for_conflicts_in_files(files: List[PreMediaContainer]) -> None:
 
         locations = {k: v for k, v in locations.items() if len(v) > 1}
 
-        # TODO: Maybe filter out duplicates by file size
-
-        # TODO: Server
+        # Later: Server
 
         for new_row in locations.values():
             for i, item in enumerate(new_row):
-                basename, ext = os.path.splitext(item.name)
-                item.name = basename + f"({i}-{len(new_row) - 1})" + ext
+                basename, ext = os.path.splitext(item._name)
+                item._name = basename + f"({i}-{len(new_row) - 1})" + ext
 
 
 class CourseDownloader:
@@ -392,25 +455,26 @@ class CourseDownloader:
         self.user = user
 
     def start(self) -> None:
+        global pre_status
+        pre_status = InfoStatus()
+        while PreStatusInfo.content.value:
+            PreStatusInfo.content.value.pop()
+
         pre_status.start()
-        pre_status.set_status(PreStatusInfo.authenticating0)
-        self.make_helper()
 
-        pre_containers = self.build_files()
-
-        check_for_conflicts_in_files(pre_containers)
-
-        if is_testing:
-            pre_containers = _course_downloader_transformation(pre_containers)
+        self.helper = RequestHelper(self.user)
 
+        pre_containers = self.helper.download_content()
         media_containers = self.make_files(pre_containers)
+
         global downloading_files
         downloading_files = media_containers
+        self.helper.make_course_paths()
 
         # Make the runner a thread in case of a user needing to exit the program → downloading is done in the main thread
         global status
         throttler = DownloadThrottler()
-        status = Status(len(media_containers), args.num_threads, throttler)
+        status = DownloadStatus(media_containers, args.num_threads, throttler)
         downloader = Thread(target=self.download_files, args=(media_containers, throttler))
 
         pre_status.stop()
@@ -420,14 +484,6 @@ class CourseDownloader:
         downloader.join()
         status.join(0)
 
-    def make_helper(self) -> None:
-        self.helper = RequestHelper(self.user)
-
-    def build_files(self) -> List[PreMediaContainer]:
-        assert self.helper is not None
-
-        return self.helper.download_content()
-
     def make_files(self, files: List[PreMediaContainer]) -> List[MediaContainer]:
         assert self.helper is not None
 
@@ -437,6 +493,8 @@ class CourseDownloader:
         return filtered_files
 
     def download_files(self, files: List[MediaContainer], throttler: DownloadThrottler) -> None:
+        exception_lock = threading.Lock()
+
         def download(file: MediaContainer) -> None:
             assert status is not None
             if enable_multithread:
@@ -445,7 +503,12 @@ class CourseDownloader:
                 thread_id = 0
 
             status.add(thread_id, file)
-            file.download(throttler)
+            try:
+                file.download(throttler)
+            except Exception:
+                with exception_lock:
+                    generate_error_message()
+
             status.finish(thread_id)
 
         if enable_multithread:
@@ -456,7 +519,7 @@ class CourseDownloader:
                 download(file)
 
     @staticmethod
-    @on_kill()
+    @on_kill(2)
     def shutdown_running_downloads(*_: Any) -> None:
         if downloading_files is None:
             return
@@ -472,6 +535,6 @@ class CourseDownloader:
             time.sleep(0.25)
 
 
-pre_status = PreStatus()
-status: Optional[Status] = None
+pre_status = InfoStatus()
+status: Optional[DownloadStatus] = None
 downloading_files: Optional[List[MediaContainer]] = None
